(* This file is part of xmlfuzzer.
 *
 * xmlfuzzer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * xmlfuzzer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with xmlfuzzer.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright 2010 Alexander Markov <apsheronets@gmail.com>
 * Center of Innovative Security Solutions
 * http://binanalysis.com *)

open ExtLib
open Tools
open Schema

exception Cannot_build

module type Options = sig
  val min_level:  int
  val max_level:  int
  val max_repeat: int
  val max_elem:   int
  val ignore_tags:  string list
  val ignore_attrs: string list
  val preferred_tags: string list
end

module Make(Opt: Options) = struct

  let counter = ref 0

  type any = {{ [ (Ocamlduce.Load.anyxml | Char)* ] }}

  let ignore_attr a =
    let comp x =
      let atom = a.at_decl.at_name in
      let _, s = Ocamlduce.Atom.get atom in
      let s = Ocamlduce.Utf8.get s in
      x = s in
    List.exists (comp) Opt.ignore_attrs

  let ignore_tag name =
    List.exists (fun x ->
      let _, s = Ocamlduce.Atom.get name in
      let s = Ocamlduce.Utf8.get s in
      x = s) Opt.ignore_tags

  let preferred_tag name =
    List.exists (fun x ->
      let _, s = Ocamlduce.Atom.get name in
      let s = Ocamlduce.Utf8.get s in
      x = s) Opt.preferred_tags

  let make_value_constraint f = function
    | Fixed s -> {{ {} }}, s
    | Default _ -> {{ {} }}, failwith "default"
    | Free -> f ()

  let make_simple_definition l sd =
    {{ {} }}, Rand.value sd.st_facets

  let make_attribute d =
    let atom = d.at_name in
    Ocamlduce.Record.make [(atom, Rand.value d.at_typdef.st_facets)]

  let make_attribute_use al =
    List.fold_left (fun (acc : Ocamlduce.Record.t ) a ->
      let ignored = ignore_attr a in
      if a.at_required && ignored
      then raise Cannot_build
      else if (a.at_required || Random.bool ()) && (not ignored)
      then
        let a = make_attribute a.at_decl in
        {{ acc ++ a }}
      else
        acc) {{ {} }} al

  let rec make_particle l p =
    match p.part_term with
    | Elt e ->
        let max = match p.part_max with
          | None -> Opt.max_repeat
          | Some i -> {: i :} in
        let min = {: {{ p.part_min }} :} in
        let count =
          if preferred_tag e.el_name && max >= 1 then Rand.int ~min:1 ~max () else
          if l > Opt.max_level && min <= 0 then 0 else
          if l < Opt.min_level && max >= 1 then Rand.int ~min:1 ~max ()
          else Rand.int ~min ~max () in
        let rec loop (acc : any) n attempts =
          if (n > count) || (!counter > Opt.max_elem)
          then acc
          else
            try
              let t = {{ [(make_element l e)] }} in
              loop {{ acc @ t }} (n+1) 1
            with
              | Cannot_build ->
                  if attempts < 5
                  then loop acc n (succ attempts)
                  else if n > min
                  then acc
                  else raise Cannot_build in
        loop {{ [] }} 1 1
    | Model (Choice   pl) ->
        let rec loop acc =
          let remove i l =
            List.rev (snd (List.fold_left (fun (n, acc) x ->
              if n = i then (succ n, acc)
              else (succ n, x :: acc)) (0, []) l)) in
          let n =
            (try
              (* try to choose preferred tag *)
              let n, _ =
                List.findi (fun _ p ->
                  match p.part_term with
                  | Elt e -> preferred_tag e.el_name
                  | _ -> false) acc in
              (* but give a small chance for another tags *)
              if Random.int 10 > 7 then raise Not_found else n
            with Not_found ->
              (try Random.int (List.length acc)
              with _ -> raise Cannot_build)) in
          let x = List.nth acc n in
          try {{ (make_particle l x) }}
          with Cannot_build -> loop (remove n acc) in
        loop pl
    | Model (All      pl) ->
        (* FIXME: I don't know what to do *)
        {{ [] }}
    | Model (Sequence pl) ->
        xmlfold
          (fun (acc : any ) x ->
            {{ acc @ (make_particle l x) }} ) {{ [] }} pl
    | Wildcard _ -> {{ [] }}

  and make_complex_definition l d : (Ocamlduce.Record.t * any ) =
    let make_content l content : any =
      match content with
      | CT_empty -> {{ [] }}
      | CT_simple sd -> snd (make_simple_definition l sd)
      | CT_model (p, _) -> make_particle l p in
    let attrs =
      make_attribute_use d.ct_attrs in
    match d.ct_deriv with
    | `Extension ->
        let extra_attrs, def = make_type_definition (l+1) d.ct_base in
        let c = make_content l d.ct_content in
        {{ attrs ++ extra_attrs }}, {{ def @ c }}
    | `Restriction ->
        attrs, make_content l d.ct_content

  and make_type_definition l d : (Ocamlduce.Record.t * any ) =
    match d with
    | Simple sd -> make_simple_definition  l sd
    | Complex d ->
        let make_content l content : any =
          match content with
          | CT_empty -> {{ [] }}
          | CT_simple sd -> snd (make_simple_definition l sd)
          | CT_model (p, _) -> make_particle l p in
        let attrs =
          make_attribute_use d.ct_attrs in
        match d.ct_deriv with
        | `Extension ->
            let extra_attrs, def = make_type_definition (l+1) d.ct_base in
            let c = make_content l d.ct_content in
            {{ attrs ++ extra_attrs }}, {{ def @ c }}
        | `Restriction ->
            attrs, make_content l d.ct_content

  and make_element l e : {{ Ocamlduce.Load.anyxml }} =
    let name = e.el_name in
    if ignore_tag name
    then raise Cannot_build
    else
      let attrs, content =
        let l = succ l in
        make_value_constraint (fun () ->
          make_type_definition l (Lazy.force e.el_typdef))
          e.el_cstr in
      incr counter;
      {{ < (name) (attrs) >content }}

  let make_element l e : {{ Ocamlduce.Load.anyxml }} =
    counter := 0;
    make_element l e

end

